/*
 * gtree: utility routines to use the GtkDTree widget in gmc.
 *
 * Miguel de Icaza
 */
#include <config.h>
#include <sys/types.h>
#include "tree.h"
#include <gnome.h>
#include "gtree.h"

void
sync_tree (char *pathname)
{
}

void
tree_check (const char *dir)
{
}      

void
start_tree_check (WTree *tree)
{
}

void
do_tree_check (WTree *tree, const char *subname)
{
}

void
end_tree_check (WTree *tree)
{
}

